type id = string
type typ = string
type op = string

type formal_args = (id * typ) list

type stmt = 
    Class of id * id * stmt list
  | Var of id * exp * typ
  | If of exp * stmt * stmt
  | While of exp * stmt
  | Block of stmt list
  | ExpStmt of exp
  | Assign of id * exp
  | Function of id * formal_args * typ * stmt
and exp =
    Literal of int
  | Fliteral of string
  | BoolLit of bool
  | Id of string
  | Call of id * exp list
  | Binop of exp * op * exp
  | Unop of op * exp
  | Return of exp


type program = stmt list 

let string_of_expr = function
    Literal(l) -> string_of_int l
  | Fliteral(l) -> l
  | BoolLit(true) -> "true"
  | BoolLit(false) -> "false"
  | Id(s) -> s



let rec string_of_exp exp =
  match exp with
  | Literal v -> string_of_int v
  | Fliteral s -> s
  | BoolLit v -> if v then "true" else "false"
  | Id id -> id
  | Call (f, arg::args) -> f ^ "(" ^ 
    List.fold_left (fun s e -> s ^ string_of_exp e) (string_of_exp arg) args
    ^ ")"
  | Call (f, []) -> f ^ "()"
  | Binop (e1, op, e2) ->
    string_of_exp e1 ^ op ^ string_of_exp e2
  | Unop (op, e) ->
    op ^ string_of_exp e
  | Return e -> "return " ^ string_of_exp e


let rec string_of_program ast =
  match ast with
  | [] -> ""
  | Class(name, super, stmts) :: rest -> 
    "class: " ^ name ^ " <: " ^ super ^ " {\n"
      ^ string_of_program stmts ^
    "\n}\n"
    ^ (string_of_program rest)
  | Var(id, exp, typ) :: rest ->
    id ^ ":" ^ typ ^ " = " ^ (string_of_exp exp)  ^
    "\n"
    ^ (string_of_program rest)
  | If(e, t, f) :: rest ->
    "if (" ^ (string_of_exp e) ^ ")\n" ^
    (string_of_program [t]) ^ "else\n" ^
    (string_of_program [f]) ^ "\n"
    ^ (string_of_program rest)
  | Block stmts :: rest ->
    "{\n" ^ string_of_program stmts ^ "}\n"
    ^ (string_of_program rest)
  | ExpStmt e :: rest ->
    string_of_exp e
    ^ "\n"
    ^ (string_of_program rest)
  | Function (name, fargs, rettyp, stmt) :: rest ->
    "function " ^ name ^ "\n"
    ^ (string_of_program rest)
  | While (e, stmt) :: rest ->
    "while (" ^ string_of_exp e ^ ") {\n" ^
      (string_of_program [stmt]) ^
    "\n}" ^ "\n"
    ^ (string_of_program rest)
  | _ -> "not impl"
